@@ -73,12 +73,10 @@ ac_table = {
             '--pull-latest',
             '--volumes',
             '--gpus',
+            '--disable-auto-volume',
             '--host',
-            '--port-jinad',
             '--quiet-remote-logs',
             '--upload-files',
-            '--runtime-backend',
-            '--runtime',
             '--runtime-cls',
             '--timeout-ready',
             '--env',
@@ -88,6 +86,8 @@ ac_table = {
             '--shards',
             '--replicas',
             '--port',
+            '--monitoring',
+            '--port-monitoring',
             '--install-requirements',
             '--force-update',
             '--force',
@@ -146,6 +146,9 @@ ac_table = {
             '--no-crud-endpoints',
             '--expose-endpoints',
             '--uvicorn-kwargs',
+            '--grpc-server-kwargs',
+            '--ssl-certfile',
+            '--ssl-keyfile',
             '--expose-graphql-endpoint',
             '--protocol',
             '--host',
@@ -154,9 +157,8 @@ ac_table = {
             '--graph-description',
             '--graph-conditions',
             '--deployments-addresses',
+            '--deployments-disable-reduce',
             '--compression',
-            '--runtime-backend',
-            '--runtime',
             '--runtime-cls',
             '--timeout-ready',
             '--env',
@@ -166,6 +168,8 @@ ac_table = {
             '--shards',
             '--replicas',
             '--port',
+            '--monitoring',
+            '--port-monitoring',
             '--uses-before-address',
             '--uses-after-address',
             '--connection-list',
@@ -190,6 +194,7 @@ ac_table = {
             '--force-update',
             '--force',
             '--secret',
+            '--no-cache',
             '--public',
             '--private',
         ],
@@ -228,12 +233,10 @@ ac_table = {
             '--pull-latest',
             '--volumes',
             '--gpus',
+            '--disable-auto-volume',
             '--host',
-            '--port-jinad',
             '--quiet-remote-logs',
             '--upload-files',
-            '--runtime-backend',
-            '--runtime',
             '--runtime-cls',
             '--timeout-ready',
             '--env',
@@ -243,6 +246,8 @@ ac_table = {
             '--shards',
             '--replicas',
             '--port',
+            '--monitoring',
+            '--port-monitoring',
             '--install-requirements',
             '--force-update',
             '--force',
@@ -278,12 +283,10 @@ ac_table = {
             '--pull-latest',
             '--volumes',
             '--gpus',
+            '--disable-auto-volume',
             '--host',
-            '--port-jinad',
             '--quiet-remote-logs',
             '--upload-files',
-            '--runtime-backend',
-            '--runtime',
             '--runtime-cls',
             '--timeout-ready',
             '--env',
@@ -293,6 +296,8 @@ ac_table = {
             '--shards',
             '--replicas',
             '--port',
+            '--monitoring',
+            '--port-monitoring',
             '--install-requirements',
             '--force-update',
             '--force',
@@ -313,7 +318,6 @@ ac_table = {
             '--proxy',
             '--port',
             '--tls',
-            '--https',
             '--asyncio',
             '--return-responses',
             '--protocol',
