@@ -1,5 +1,6 @@
 import asyncio
 import functools
+import inspect
 import json
 import math
 import os
@@ -10,24 +11,33 @@ import threading
 import time
 import uuid
 import warnings
-
 from argparse import ArgumentParser, Namespace
+from collections.abc import MutableMapping
 from datetime import datetime
 from itertools import islice
+from socket import AF_INET, SOCK_STREAM, socket
 from types import SimpleNamespace
 from typing import (
-    Tuple,
-    Optional,
-    Iterator,
+    TYPE_CHECKING,
     Any,
-    Union,
-    List,
+    Callable,
     Dict,
-    Set,
-    Sequence,
     Iterable,
+    Iterator,
+    List,
+    Optional,
+    Sequence,
+    Set,
+    Tuple,
+    TypeVar,
+    Union,
 )
 
+from packaging import version as pckg_version
+from rich.console import Console
+
+from jina import __docarray_version__, __windows__
+
 __all__ = [
     'batch_iterator',
     'parse_arg',
@@ -45,13 +55,22 @@ __all__ = [
     'convert_tuple_to_list',
     'run_async',
     'deprecated_alias',
+    'retry',
     'countdown',
     'CatchAllCleanupContextManager',
     'download_mermaid_url',
     'get_readable_size',
     'get_or_reuse_loop',
+    'T',
+    'get_rich_console',
+    'docarray_graphql_compatible',
 ]
 
+if TYPE_CHECKING:
+    from docarray import DocumentArray
+
+T = TypeVar('T')
+
 
 def deprecated_alias(**aliases):
     """
@@ -62,12 +81,20 @@ def deprecated_alias(**aliases):
     For example:
         .. highlight:: python
         .. code-block:: python
-            @deprecated_alias(input_fn=('inputs', 0), buffer=('input_fn', 0), callback=('on_done', 1), output_fn=('on_done', 1))
+
+            @deprecated_alias(
+                input_fn=('inputs', 0),
+                buffer=('input_fn', 0),
+                callback=('on_done', 1),
+                output_fn=('on_done', 1),
+            )
+            def some_function(inputs, input_fn, on_done):
+                pass
 
     :param aliases: maps aliases to new arguments
     :return: wrapper
     """
-    from .excepts import NotSupportedError
+    from jina.excepts import NotSupportedError
 
     def _rename_kwargs(func_name: str, kwargs, aliases):
         """
@@ -125,6 +152,57 @@ def deprecated_alias(**aliases):
     return deco
 
 
+def deprecated_method(new_function_name):
+    def deco(func):
+        def wrapper(*args, **kwargs):
+            warnings.warn(
+                f'`{func.__name__}` is renamed to `{new_function_name}`, the usage of `{func.__name__}` is '
+                f'deprecated and will be removed.',
+                DeprecationWarning,
+            )
+            return func(*args, **kwargs)
+
+        return wrapper
+
+    return deco
+
+
+def retry(
+    num_retry: int = 3,
+    message: str = 'Calling {func_name} failed, retry attempt {attempt}/{num_retry}. Error: {error!r}',
+):
+    """
+    Retry calling a function again in case of an error.
+
+    :param num_retry: number of times to retry
+    :param message: message to log when error happened
+    :return: wrapper
+    """
+    from jina.logging.predefined import default_logger
+
+    def decorator(func):
+        @functools.wraps(func)
+        def wrapper(*args, **kwargs):
+            for i in range(num_retry):
+                try:
+                    return func(*args, **kwargs)
+                except Exception as e:
+                    default_logger.warning(
+                        message.format(
+                            func_name=func.__name__,
+                            attempt=i + 1,
+                            num_retry=num_retry,
+                            error=e,
+                        )
+                    )
+                    if i + 1 == num_retry:
+                        raise
+
+        return wrapper
+
+    return decorator
+
+
 def get_readable_size(num_bytes: Union[int, float]) -> str:
     """
     Transform the bytes into readable value with different units (e.g. 1 KB, 20 MB, 30.1 GB).
@@ -135,9 +213,9 @@ def get_readable_size(num_bytes: Union[int, float]) -> str:
     num_bytes = int(num_bytes)
     if num_bytes < 1024:
         return f'{num_bytes} Bytes'
-    elif num_bytes < 1024 ** 2:
+    elif num_bytes < 1024**2:
         return f'{num_bytes / 1024:.1f} KB'
-    elif num_bytes < 1024 ** 3:
+    elif num_bytes < 1024**3:
         return f'{num_bytes / (1024 ** 2):.1f} MB'
     else:
         return f'{num_bytes / (1024 ** 3):.1f} GB'
@@ -154,8 +232,9 @@ def batch_iterator(
     For example:
     .. highlight:: python
     .. code-block:: python
+
             for req in batch_iterator(data, batch_size, split_over_axis):
-                # Do something with batch
+                pass  # Do something with batch
 
     :param data: Data source.
     :param batch_size: Size of one batch.
@@ -187,8 +266,9 @@ def batch_iterator(
             yield data[_ : _ + batch_size]
     elif isinstance(data, Iterable):
         # as iterator, there is no way to know the length of it
+        iterator = iter(data)
         while True:
-            chunk = tuple(islice(data, batch_size))
+            chunk = tuple(islice(iterator, batch_size))
             if not chunk:
                 return
             yield chunk
@@ -237,7 +317,9 @@ def countdown(t: int, reason: str = 'I am blocking this thread') -> None:
     For example:
         .. highlight:: python
         .. code-block:: python
-            countdown(10, reason=colored('re-fetch access token', 'cyan', attrs=['bold', 'reverse']))
+            countdown(
+                10, reason=colored('re-fetch access token', 'cyan', attrs=['bold', 'reverse'])
+            )
 
     :param t: Countdown time.
     :param reason: A string message of reason for this Countdown.
@@ -374,46 +456,70 @@ def random_name() -> str:
     return '_'.join(random.choice(_random_names[j]) for j in range(2))
 
 
+assigned_ports = set()
+unassigned_ports = []
+DEFAULT_MIN_PORT = 49153
+MAX_PORT = 65535
+
+
+def reset_ports():
+    def _get_unassigned_ports():
+        # if we are running out of ports, lower default minimum port
+        if MAX_PORT - DEFAULT_MIN_PORT - len(assigned_ports) < 100:
+            min_port = int(os.environ.get('JINA_RANDOM_PORT_MIN', '16384'))
+        else:
+            min_port = int(
+                os.environ.get('JINA_RANDOM_PORT_MIN', str(DEFAULT_MIN_PORT))
+            )
+        max_port = int(os.environ.get('JINA_RANDOM_PORT_MAX', str(MAX_PORT)))
+        return set(range(min_port, max_port + 1)) - set(assigned_ports)
+
+    unassigned_ports.clear()
+    assigned_ports.clear()
+    unassigned_ports.extend(_get_unassigned_ports())
+    random.shuffle(unassigned_ports)
+
+
 def random_port() -> Optional[int]:
     """
-    Get a random available port number from '49153' to '65535'.
+    Get a random available port number.
 
     :return: A random port.
     """
 
-    import threading
-    import multiprocessing
-    from contextlib import closing
-    import socket
+    def _random_port():
+        import socket
 
-    def _get_port(port=0):
-        with multiprocessing.Lock():
-            with threading.Lock():
-                with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
-                    try:
-                        s.bind(('', port))
-                        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
-                        return s.getsockname()[1]
-                    except OSError:
-                        pass
+        def _check_bind(port):
+            with socket.socket() as s:
+                try:
+                    s.bind(('', port))
+                    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
+                    return port
+                except OSError:
+                    return None
 
-    _port = None
-    if 'JINA_RANDOM_PORT_MIN' in os.environ or 'JINA_RANDOM_PORT_MAX' in os.environ:
-        min_port = int(os.environ.get('JINA_RANDOM_PORT_MIN', '49153'))
-        max_port = int(os.environ.get('JINA_RANDOM_PORT_MAX', '65535'))
-        all_ports = list(range(min_port, max_port + 1))
-        random.shuffle(all_ports)
-        for _port in all_ports:
-            if _get_port(_port) is not None:
+        _port = None
+        if len(unassigned_ports) == 0:
+            reset_ports()
+        for idx, _port in enumerate(unassigned_ports):
+            if _check_bind(_port) is not None:
                 break
         else:
             raise OSError(
-                f'can not find an available port between [{min_port}, {max_port}].'
+                f'can not find an available port in {len(unassigned_ports)} unassigned ports, assigned already {len(assigned_ports)} ports'
             )
-    else:
-        _port = _get_port()
+        int_port = int(_port)
+        unassigned_ports.pop(idx)
+        assigned_ports.add(int_port)
+        return int_port
 
-    return int(_port)
+    try:
+        return _random_port()
+    except OSError:
+        assigned_ports.clear()
+        unassigned_ports.clear()
+        return _random_port()
 
 
 def random_identity(use_uuid1: bool = False) -> str:
@@ -431,7 +537,7 @@ def random_identity(use_uuid1: bool = False) -> str:
     :return: A random UUID.
 
     """
-    return str(random_uuid(use_uuid1))
+    return random_uuid(use_uuid1).hex
 
 
 def random_uuid(use_uuid1: bool = False) -> uuid.UUID:
@@ -555,7 +661,7 @@ _COLORS = {
 
 _RESET = '\033[0m'
 
-if os.name == 'nt':
+if __windows__:
     os.system('color')
 
 
@@ -619,23 +725,69 @@ def colored(
     return text
 
 
-class ColorContext:
-    def __init__(self, color: str, bold: Optional[bool] = False):
-        self._color = color
-        self._bold = bold
+def colored_rich(
+    text: str,
+    color: Optional[str] = None,
+    on_color: Optional[str] = None,
+    attrs: Optional[Union[str, list]] = None,
+) -> str:
+    """
+    Give the text with color. You should only use it when printing with rich print. Othersiwe please see the colored
+    function
 
-    def __enter__(self):
-        if self._bold:
-            fmt_str = '\033[1;%dm'
-        else:
-            fmt_str = '\033[0;%dm'
+    :param text: The target text
+    :param color: The color of text
+    :param on_color: The on color of text: ex on yellow
+    :param attrs: Attributes of color
 
-        c = fmt_str % (_COLORS[self._color])
-        print(c, flush=True, end='')
-        return self
+    :return: Colored text.
+    """
+    if 'JINA_LOG_NO_COLOR' not in os.environ:
+        if color:
+            text = _wrap_text_in_rich_bracket(text, color)
+        if on_color:
+            text = _wrap_text_in_rich_bracket(text, on_color)
 
-    def __exit__(self, typ, value, traceback):
-        print(_RESET, flush=True, end='')
+        if attrs:
+            if isinstance(attrs, str):
+                attrs = [attrs]
+            if isinstance(attrs, list):
+                for attr in attrs:
+                    text = _wrap_text_in_rich_bracket(text, attr)
+    return text
+
+
+def _wrap_text_in_rich_bracket(text: str, wrapper: str):
+    return f'[{wrapper}]{text}[/{wrapper}]'
+
+
+def warn_unknown_args(unknown_args: List[str]):
+    """Creates warnings for all given arguments.
+
+    :param unknown_args: arguments that are possibly unknown to Jina
+    """
+
+    from cli.lookup import _build_lookup_table
+
+    all_args = _build_lookup_table()[0]
+    has_migration_tip = False
+    real_unknown_args = []
+    warn_strs = []
+    for arg in unknown_args:
+        if arg.replace('--', '') not in all_args:
+            from jina.parsers.deprecated import get_deprecated_replacement
+
+            new_arg = get_deprecated_replacement(arg)
+            if new_arg:
+                if not has_migration_tip:
+                    warn_strs.append('Migration tips:')
+                    has_migration_tip = True
+                warn_strs.append(f'\t`{arg}` has been renamed to `{new_arg}`')
+            real_unknown_args.append(arg)
+
+    if real_unknown_args:
+        warn_strs = [f'ignored unknown argument: {real_unknown_args}.'] + warn_strs
+        warnings.warn(''.join(warn_strs))
 
 
 class ArgNamespace:
@@ -650,7 +802,7 @@ class ArgNamespace:
         :return: argument list
         """
         args = []
-        from .executors import BaseExecutor
+        from jina.serve.executors import BaseExecutor
 
         for k, v in kwargs.items():
             k = k.replace('_', '-')
@@ -670,23 +822,38 @@ class ArgNamespace:
 
     @staticmethod
     def kwargs2namespace(
-        kwargs: Dict[str, Union[str, int, bool]], parser: ArgumentParser
+        kwargs: Dict[str, Union[str, int, bool]],
+        parser: ArgumentParser,
+        warn_unknown: bool = False,
+        fallback_parsers: Optional[List[ArgumentParser]] = None,
+        positional_args: Optional[Tuple[str, ...]] = None,
     ) -> Namespace:
         """
         Convert dict to a namespace.
 
         :param kwargs: dictionary of key-values to be converted
         :param parser: the parser for building kwargs into a namespace
+        :param warn_unknown: True, if unknown arguments should be logged
+        :param fallback_parsers: a list of parsers to help resolving the args
+        :param positional_args: some parser requires positional arguments to be presented
         :return: argument list
         """
         args = ArgNamespace.kwargs2list(kwargs)
-        try:
-            p_args, unknown_args = parser.parse_known_args(args)
-        except SystemExit:
-            raise ValueError(
-                f'bad arguments "{args}" with parser {parser}, '
-                'you may want to double check your args '
-            )
+        if positional_args:
+            args += positional_args
+        p_args, unknown_args = parser.parse_known_args(args)
+        unknown_args = list(filter(lambda x: x.startswith('--'), unknown_args))
+        if warn_unknown and unknown_args:
+            _leftovers = set(unknown_args)
+            if fallback_parsers:
+                for p in fallback_parsers:
+                    _, _unk_args = p.parse_known_args(args)
+                    _leftovers = _leftovers.intersection(_unk_args)
+                    if not _leftovers:
+                        # all args have been resolved
+                        break
+            warn_unknown_args(_leftovers)
+
         return p_args
 
     @staticmethod
@@ -717,20 +884,20 @@ class ArgNamespace:
         """Convert argparse.Namespace to dict to be uploaded via REST.
 
         :param args: namespace or dict or namespace to dict.
-        :return: pea args
+        :return: pod args
         """
         if isinstance(args, Namespace):
             return vars(args)
         elif isinstance(args, dict):
-            pea_args = {}
+            pod_args = {}
             for k, v in args.items():
                 if isinstance(v, Namespace):
-                    pea_args[k] = vars(v)
+                    pod_args[k] = vars(v)
                 elif isinstance(v, list):
-                    pea_args[k] = [vars(_) for _ in v]
+                    pod_args[k] = [vars(_) for _ in v]
                 else:
-                    pea_args[k] = v
-            return pea_args
+                    pod_args[k] = v
+            return pod_args
 
 
 def is_valid_local_config_source(path: str) -> bool:
@@ -742,7 +909,7 @@ def is_valid_local_config_source(path: str) -> bool:
     :return: True if the path is valid else False.
     """
     try:
-        from .jaml import parse_config_source
+        from jina.jaml import parse_config_source
 
         parse_config_source(path)
         return True
@@ -756,27 +923,33 @@ def get_full_version() -> Optional[Tuple[Dict, Dict]]:
 
     :return: Version information and environment variables
     """
-    import os, grpc, zmq, numpy, google.protobuf, yaml, platform
-    from . import (
-        __version__,
-        __proto_version__,
-        __jina_env__,
-        __uptime__,
-        __unset_msg__,
-    )
+    import os
+    import platform
+    from uuid import getnode
+
+    import google.protobuf
+    import grpc
+    import yaml
     from google.protobuf.internal import api_implementation
     from grpc import _grpcio_metadata
+
+    from jina import (
+        __docarray_version__,
+        __jina_env__,
+        __proto_version__,
+        __unset_msg__,
+        __uptime__,
+        __version__,
+    )
     from jina.logging.predefined import default_logger
-    from uuid import getnode
 
     try:
 
         info = {
             'jina': __version__,
+            'docarray': __docarray_version__,
             'jina-proto': __proto_version__,
             'jina-vcs-tag': os.environ.get('JINA_VCS_VERSION', __unset_msg__),
-            'libzmq': zmq.zmq_version(),
-            'pyzmq': numpy.__version__,
             'protobuf': google.protobuf.__version__,
             'proto-backend': api_implementation._default_implementation_type,
             'grpcio': getattr(grpc, '__version__', _grpcio_metadata.__version__),
@@ -815,17 +988,21 @@ def format_full_version_info(info: Dict, env_info: Dict) -> str:
     return version_info + '\n' + env_info
 
 
-def _use_uvloop():
-    if 'JINA_DISABLE_UVLOOP' in os.environ:
-        return
-    try:
-        import uvloop
+def _update_policy():
+    if __windows__:
+        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
 
-        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
-    except ModuleNotFoundError:
-        warnings.warn(
-            'Install `uvloop` via `pip install "jina[uvloop]"` for better performance.'
-        )
+    elif 'JINA_DISABLE_UVLOOP' in os.environ:
+        return
+    else:
+        try:
+            import uvloop
+
+            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
+        except ModuleNotFoundError:
+            warnings.warn(
+                'Install `uvloop` via `pip install "jina[uvloop]"` for better performance.'
+            )
 
 
 def get_or_reuse_loop():
@@ -839,7 +1016,7 @@ def get_or_reuse_loop():
         if loop.is_closed():
             raise RuntimeError
     except RuntimeError:
-        _use_uvloop()
+        _update_policy()
         # no running event loop
         # create a new loop
         loop = asyncio.new_event_loop()
@@ -878,7 +1055,7 @@ class CatchAllCleanupContextManager:
         pass
 
     def __exit__(self, exc_type, exc_val, exc_tb):
-        if exc_type is not None:
+        if exc_type:
             self.sub_context.__exit__(exc_type, exc_val, exc_tb)
 
 
@@ -1018,7 +1195,7 @@ def get_public_ip(timeout: float = 0.3):
     :return: Public IP address.
 
     .. warn::
-        Set :param:`timeout` to a large number will block the Flow.
+        Set `timeout` to a large number will block the Flow.
 
     """
     import urllib.request
@@ -1029,7 +1206,7 @@ def get_public_ip(timeout: float = 0.3):
         try:
             req = urllib.request.Request(url, headers={'User-Agent': 'Mozilla/5.0'})
             with urllib.request.urlopen(req, timeout=timeout) as fp:
-                _ip = fp.read().decode()
+                _ip = fp.read().decode().strip()
                 results.append(_ip)
 
         except:
@@ -1090,6 +1267,10 @@ def is_jupyter() -> bool:  # pragma: no cover
         return False  # Other type (?)
 
 
+def iscoroutinefunction(func: Callable):
+    return inspect.iscoroutinefunction(func)
+
+
 def run_async(func, *args, **kwargs):
     """Generalized asyncio.run for jupyter notebook.
 
@@ -1100,12 +1281,16 @@ def run_async(func, *args, **kwargs):
     .. see_also:
         https://stackoverflow.com/questions/55409641/asyncio-run-cannot-be-called-from-a-running-event-loop
 
+    call `run_async(my_function, any_event_loop=True, *args, **kwargs)` to enable run with any eventloop
+
     :param func: function to run
     :param args: parameters
     :param kwargs: key-value parameters
     :return: asyncio.run(func)
     """
 
+    any_event_loop = kwargs.pop('any_event_loop', False)
+
     class _RunThread(threading.Thread):
         """Create a running thread when in Jupyter notebook."""
 
@@ -1121,14 +1306,14 @@ def run_async(func, *args, **kwargs):
     if loop and loop.is_running():
         # eventloop already exist
         # running inside Jupyter
-        if is_jupyter():
+        if any_event_loop or is_jupyter():
             thread = _RunThread()
             thread.start()
             thread.join()
             try:
                 return thread.result
             except AttributeError:
-                from .excepts import BadClient
+                from jina.excepts import BadClient
 
                 raise BadClient(
                     'something wrong when running the eventloop, result can not be retrieved'
@@ -1138,11 +1323,11 @@ def run_async(func, *args, **kwargs):
             raise RuntimeError(
                 'you have an eventloop running but not using Jupyter/ipython, '
                 'this may mean you are using Jina with other integration? if so, then you '
-                'may want to use Clien/Flow(asyncio=True). If not, then '
+                'may want to use Client/Flow(asyncio=True). If not, then '
                 'please report this issue here: https://github.com/jina-ai/jina'
             )
     else:
-        return asyncio.run(func(*args, **kwargs))
+        return get_or_reuse_loop().run_until_complete(func(*args, **kwargs))
 
 
 def slugify(value):
@@ -1163,7 +1348,10 @@ def is_yaml_filepath(val) -> bool:
     :param val: Path of target file.
     :return: True if the file is YAML else False.
     """
-    r = r'^[/\w\-\_\.]+.ya?ml$'
+    if __windows__:
+        r = r'.*.ya?ml$'  # TODO: might not be exhaustive
+    else:
+        r = r'^[/\w\-\_\.]+.ya?ml$'
     return re.match(r, val.strip()) is not None
 
 
@@ -1194,8 +1382,10 @@ def find_request_binding(target):
     :param target: the target class to check
     :return: a dictionary with key as request type and value as method name
     """
-    import ast, inspect
-    from . import __default_endpoint__
+    import ast
+    import inspect
+
+    from jina import __default_endpoint__
 
     res = {}
 
@@ -1245,26 +1435,24 @@ def dunder_get(_dict: Any, key: str) -> Any:
     except ValueError:
         pass
 
-    from google.protobuf.struct_pb2 import ListValue
-    from google.protobuf.struct_pb2 import Struct
-    from google.protobuf.pyext._message import MessageMapContainer
+    from google.protobuf.struct_pb2 import ListValue, Struct
 
     if isinstance(part1, int):
         result = _dict[part1]
-    elif isinstance(_dict, (Iterable, ListValue)):
-        result = _dict[part1]
-    elif isinstance(_dict, (dict, Struct, MessageMapContainer)):
+    elif isinstance(_dict, (dict, Struct, MutableMapping)):
         if part1 in _dict:
             result = _dict[part1]
         else:
             result = None
+    elif isinstance(_dict, (Iterable, ListValue)):
+        result = _dict[part1]
     else:
         result = getattr(_dict, part1)
 
     return dunder_get(result, part2) if part2 else result
 
 
-if False:
+if TYPE_CHECKING:
     from fastapi import FastAPI
 
 
@@ -1278,7 +1466,6 @@ def extend_rest_interface(app: 'FastAPI') -> 'FastAPI':
     .. code-block:: python
 
         def extend_rest_interface(app: 'FastAPI'):
-
             @app.get('/extension1')
             async def root():
                 return {"message": "Hello World"}
@@ -1304,3 +1491,141 @@ def get_ci_vendor() -> Optional[str]:
                 for k in c['env']:
                     if k in os.environ:
                         return c['constant']
+
+
+def deprecate_by(new_fn):
+    def _f(*args, **kwargs):
+        import inspect
+
+        old_fn_name = inspect.stack()[1][4][0].strip().split("=")[0].strip()
+        warnings.warn(
+            f'`{old_fn_name}` is renamed to `{new_fn.__name__}` with the same usage, please use the latter instead. '
+            f'The old function will be removed soon.',
+            DeprecationWarning,
+        )
+        return new_fn(*args, **kwargs)
+
+    return _f
+
+
+def get_request_header() -> Dict:
+    """Return the header of request.
+
+    :return: request header
+    """
+    metas, envs = get_full_version()
+
+    header = {
+        **{f'jinameta-{k}': str(v) for k, v in metas.items()},
+        **envs,
+    }
+    return header
+
+
+def get_rich_console():
+    """
+    Function to get jina rich default console.
+    :return: rich console
+    """
+    return Console(
+        force_terminal=True,
+    )  # It forces render in any terminal, especially in PyCharm
+
+
+GRAPHQL_MIN_DOCARRAY_VERSION = '0.8.8'  # graphql requires this or higher
+
+
+def docarray_graphql_compatible():
+    """Check if installed docarray version is compatible with GraphQL features.
+
+    :return: True if compatible, False if not
+    """
+    installed_version = pckg_version.parse(__docarray_version__)
+    min_version = pckg_version.parse(GRAPHQL_MIN_DOCARRAY_VERSION)
+    return installed_version >= min_version
+
+
+from jina.helper import ArgNamespace
+from jina.parsers import set_client_cli_parser
+
+
+def parse_client(kwargs):
+    kwargs = _parse_kwargs(kwargs)
+    return ArgNamespace.kwargs2namespace(
+        kwargs, set_client_cli_parser(), warn_unknown=True
+    )
+
+
+def _parse_kwargs(kwargs: Dict[str, Any]) -> Dict[str, Any]:
+
+    if 'host' in kwargs.keys():
+        return_scheme = dict()
+        (
+            kwargs['host'],
+            return_scheme['port'],
+            return_scheme['protocol'],
+            return_scheme['tls'],
+        ) = _parse_host_scheme(kwargs['host'])
+
+        for key, value in return_scheme.items():
+            if value:
+                if key in kwargs:
+                    raise ValueError(
+                        f"You can't have two definitions of {key}: you have one in the host scheme and one in the keyword argument"
+                    )
+                elif value:
+                    kwargs[key] = value
+
+    kwargs = _add_default_port_tls(kwargs)
+    kwargs = _delete_host_slash(kwargs)
+
+    return kwargs
+
+
+def _add_default_port_tls(kwargs: Dict[str, Any]) -> Dict[str, Any]:
+    if ('tls' in kwargs) and ('port' not in kwargs):
+        kwargs['port'] = 443
+    return kwargs
+
+
+def _delete_host_slash(kwargs: Dict[str, Any]) -> Dict[str, Any]:
+    if 'host' in kwargs:
+        if kwargs['host'][-1] == '/':
+            kwargs['host'] = kwargs['host'][:-1]
+    return kwargs
+
+
+def _parse_host_scheme(host: str) -> Tuple[str, str, str, bool]:
+    scheme, _hostname, port = _parse_url(host)
+
+    tls = None
+    if scheme in ('grpcs', 'https', 'wss'):
+        scheme = scheme[:-1]
+        tls = True
+
+    if scheme == 'ws':
+        scheme = 'websocket'
+
+    return _hostname, port, scheme, tls
+
+
+def _parse_url(host):
+    if '://' in host:
+        scheme, host = host.split('://')
+    else:
+        scheme = None
+
+    if ':' in host:
+        host, port = host.split(':')
+    else:
+        port = None
+
+    return scheme, host, port
+
+
+def is_port_free(host, port):
+    with socket(AF_INET, SOCK_STREAM) as session:
+        if session.connect_ex((host, port)) == 0:
+            return False
+        else:
+            return True
