@@ -2,18 +2,14 @@ import inspect
 import multiprocessing
 import os
 import threading
-from concurrent.futures import ThreadPoolExecutor
+import warnings
 from types import SimpleNamespace
 from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union
 
 from jina import __args_executor_init__, __default_endpoint__
-from jina.helper import (
-    ArgNamespace,
-    T,
-    iscoroutinefunction,
-    run_in_threadpool,
-    typename,
-)
+from jina.enums import BetterEnum
+from jina.helper import ArgNamespace, T, iscoroutinefunction, typename
+from jina.importer import ImportExtensions
 from jina.jaml import JAML, JAMLCompatible, env_var_regex, internal_var_regex
 from jina.serve.executors.decorators import requests, store_init_kwargs, wrap_func
 
@@ -106,10 +102,10 @@ class BaseExecutor(JAMLCompatible, metaclass=ExecutorType):
         :param runtime_args: a dict of arguments injected from :class:`Runtime` during runtime
         :param kwargs: additional extra keyword arguments to avoid failing when extra params ara passed that are not expected
         """
-        self._thread_pool = ThreadPoolExecutor(max_workers=1)
         self._add_metas(metas)
         self._add_requests(requests)
         self._add_runtime_args(runtime_args)
+        self._init_monitoring()
 
     def _add_runtime_args(self, _runtime_args: Optional[Dict]):
         if _runtime_args:
@@ -117,6 +113,27 @@ class BaseExecutor(JAMLCompatible, metaclass=ExecutorType):
         else:
             self.runtime_args = SimpleNamespace()
 
+    def _init_monitoring(self):
+        if (
+            hasattr(self.runtime_args, 'metrics_registry')
+            and self.runtime_args.metrics_registry
+        ):
+            with ImportExtensions(
+                required=True,
+                help_text='You need to install the `prometheus_client` to use the montitoring functionality of jina',
+            ):
+                from prometheus_client import Summary
+
+            self._summary_method = Summary(
+                'process_request_seconds',
+                'Time spent when calling the executor request method',
+                registry=self.runtime_args.metrics_registry,
+                namespace='jina',
+                labelnames=('method', 'executor'),
+            )
+        else:
+            self._summary_method = None
+
     def _add_requests(self, _requests: Optional[Dict]):
         if not hasattr(self, 'requests'):
             self.requests = {}
@@ -153,6 +170,14 @@ class BaseExecutor(JAMLCompatible, metaclass=ExecutorType):
         target = SimpleNamespace()
         # set self values filtered by those non-exist, and non-expandable
         for k, v in tmp.items():
+            if k == 'workspace' and not (v is None or v == ''):
+                warnings.warn(
+                    'Setting `workspace` via `metas.workspace` is deprecated. '
+                    'Instead, use `f.add(..., workspace=...)` when defining a a Flow in Python; '
+                    'the `workspace` parameter when defining a Flow using YAML; '
+                    'or `--workspace` when starting an Executor using the CLI.',
+                    category=DeprecationWarning,
+                )
             if not hasattr(target, k):
                 if isinstance(v, str):
                     if not env_var_regex.findall(v):
@@ -228,7 +253,7 @@ class BaseExecutor(JAMLCompatible, metaclass=ExecutorType):
         if iscoroutinefunction(func):
             return await func(self, **kwargs)
         else:
-            return await run_in_threadpool(func, self._thread_pool, self, **kwargs)
+            return func(self, **kwargs)
 
     @property
     def workspace(self) -> Optional[str]:
@@ -237,8 +262,10 @@ class BaseExecutor(JAMLCompatible, metaclass=ExecutorType):
 
         :return: returns the workspace of the current shard of this Executor.
         """
-        workspace = getattr(self.metas, 'workspace') or getattr(
-            self.runtime_args, 'workspace', None
+        workspace = (
+            getattr(self.runtime_args, 'workspace', None)
+            or getattr(self.metas, 'workspace')
+            or os.environ.get('JINA_DEFAULT_WORKSPACE_BASE')
         )
         if workspace:
             complete_workspace = os.path.join(workspace, self.metas.name)
@@ -347,6 +374,96 @@ class BaseExecutor(JAMLCompatible, metaclass=ExecutorType):
         with f:
             f.block(stop_event)
 
+    class StandaloneExecutorType(BetterEnum):
+        """
+        Type of standalone Executors
+        """
+
+        EXTERNAL = 0  # served by a gateway
+        SHARED = 1  # not served by a gateway, served by head/worker
+
+    @staticmethod
+    def to_k8s_yaml(
+        uses: str,
+        output_base_path: str,
+        k8s_namespace: Optional[str] = None,
+        executor_type: Optional[
+            StandaloneExecutorType
+        ] = StandaloneExecutorType.EXTERNAL,
+        uses_with: Optional[Dict] = None,
+        uses_metas: Optional[Dict] = None,
+        uses_requests: Optional[Dict] = None,
+        **kwargs,
+    ):
+        """
+        Converts the Executor into a set of yaml deployments to deploy in Kubernetes.
+
+        If you don't want to rebuild image on Jina Hub,
+        you can set `JINA_HUB_NO_IMAGE_REBUILD` environment variable.
+
+        :param uses: the Executor to use. Has to be containerized and accessible from K8s
+        :param output_base_path: The base path where to dump all the yaml files
+        :param k8s_namespace: The name of the k8s namespace to set for the configurations. If None, the name of the Flow will be used.
+        :param executor_type: The type of Executor. Can be external or shared. External Executors include the Gateway. Shared Executors don't. Defaults to External
+        :param uses_with: dictionary of parameters to overwrite from the default config's with field
+        :param uses_metas: dictionary of parameters to overwrite from the default config's metas field
+        :param uses_requests: dictionary of parameters to overwrite from the default config's requests field
+        :param kwargs: other kwargs accepted by the Flow, full list can be found `here <https://docs.jina.ai/api/jina.orchestrate.flow.base/>`
+        """
+        from jina import Flow
+
+        f = Flow(**kwargs).add(
+            uses=uses,
+            uses_with=uses_with,
+            uses_metas=uses_metas,
+            uses_requests=uses_requests,
+        )
+        f.to_k8s_yaml(
+            output_base_path=output_base_path,
+            k8s_namespace=k8s_namespace,
+            include_gateway=executor_type
+            == BaseExecutor.StandaloneExecutorType.EXTERNAL,
+        )
+
+    @staticmethod
+    def to_docker_compose_yaml(
+        uses: str,
+        output_path: Optional[str] = None,
+        network_name: Optional[str] = None,
+        executor_type: Optional[
+            StandaloneExecutorType
+        ] = StandaloneExecutorType.EXTERNAL,
+        uses_with: Optional[Dict] = None,
+        uses_metas: Optional[Dict] = None,
+        uses_requests: Optional[Dict] = None,
+        **kwargs,
+    ):
+        """
+        Converts the Executor into a yaml file to run with `docker-compose up`
+        :param uses: the Executor to use. Has to be containerized
+        :param output_path: The output path for the yaml file
+        :param network_name: The name of the network that will be used by the deployment name
+        :param executor_type: The type of Executor. Can be external or shared. External Executors include the Gateway. Shared Executors don't. Defaults to External
+        :param uses_with: dictionary of parameters to overwrite from the default config's with field
+        :param uses_metas: dictionary of parameters to overwrite from the default config's metas field
+        :param uses_requests: dictionary of parameters to overwrite from the default config's requests field
+        :param kwargs: other kwargs accepted by the Flow, full list can be found `here <https://docs.jina.ai/api/jina.orchestrate.flow.base/>`
+        """
+        from jina import Flow
+
+        f = Flow(**kwargs).add(
+            uses=uses,
+            uses_with=uses_with,
+            uses_metas=uses_metas,
+            uses_requests=uses_requests,
+        )
+        f.to_docker_compose_yaml(
+            output_path=output_path,
+            network_name=network_name,
+            include_gateway=executor_type
+            == BaseExecutor.StandaloneExecutorType.EXTERNAL,
+        )
+
 
 class ReducerExecutor(BaseExecutor):
     """
