@@ -1,5 +1,6 @@
 import argparse
 import asyncio
+import contextlib
 import json
 import multiprocessing
 import os
@@ -9,9 +10,11 @@ from collections import defaultdict
 from typing import Dict, List, Optional, Tuple, Union
 
 import grpc
+from grpc_reflection.v1alpha import reflection
 
 from jina.enums import PollingType
-from jina.proto import jina_pb2_grpc
+from jina.importer import ImportExtensions
+from jina.proto import jina_pb2, jina_pb2_grpc
 from jina.serve.networking import GrpcConnectionPool
 from jina.serve.runtimes.asyncio import AsyncNewLoopRuntime
 from jina.serve.runtimes.request_handlers.data_request_handler import DataRequestHandler
@@ -29,25 +32,40 @@ class HeadRuntime(AsyncNewLoopRuntime, ABC):
     def __init__(
         self,
         args: argparse.Namespace,
-        cancel_event: Optional[
-            Union['asyncio.Event', 'multiprocessing.Event', 'threading.Event']
-        ] = None,
         **kwargs,
     ):
         """Initialize grpc server for the head runtime.
         :param args: args from CLI
-        :param cancel_event: the cancel event used to wait for canceling
         :param kwargs: keyword args
         """
-        super().__init__(args, cancel_event, **kwargs)
+        super().__init__(args, **kwargs)
 
         if args.name is None:
             args.name = ''
         self.name = args.name
         self._deployment_name = os.getenv('JINA_DEPLOYMENT_NAME', 'worker')
         self.connection_pool = GrpcConnectionPool(
-            logger=self.logger, compression=args.compression
+            logger=self.logger,
+            compression=args.compression,
+            metrics_registry=self.metrics_registry,
         )
+
+        if self.metrics_registry:
+            with ImportExtensions(
+                required=True,
+                help_text='You need to install the `prometheus_client` to use the montitoring functionality of jina',
+            ):
+                from prometheus_client import Summary
+
+            self._summary = Summary(
+                'receiving_request_seconds',
+                'Time spent processing request',
+                registry=self.metrics_registry,
+                namespace='jina',
+            ).time()
+        else:
+            self._summary = contextlib.nullcontext()
+
         polling = getattr(args, 'polling', self.DEFAULT_POLLING.name)
         try:
             # try loading the polling args as json
@@ -112,7 +130,7 @@ class HeadRuntime(AsyncNewLoopRuntime, ABC):
         )
 
     async def async_setup(self):
-        """ Wait for the GRPC server to start """
+        """Wait for the GRPC server to start"""
         self._grpc_server = grpc.aio.server(
             options=[
                 ('grpc.max_send_message_length', -1),
@@ -127,13 +145,21 @@ class HeadRuntime(AsyncNewLoopRuntime, ABC):
         jina_pb2_grpc.add_JinaControlRequestRPCServicer_to_server(
             self, self._grpc_server
         )
+        service_names = (
+            jina_pb2.DESCRIPTOR.services_by_name['JinaSingleDataRequestRPC'].full_name,
+            jina_pb2.DESCRIPTOR.services_by_name['JinaDataRequestRPC'].full_name,
+            jina_pb2.DESCRIPTOR.services_by_name['JinaControlRequestRPC'].full_name,
+            reflection.SERVICE_NAME,
+        )
+        reflection.enable_server_reflection(service_names, self._grpc_server)
+
         bind_addr = f'0.0.0.0:{self.args.port}'
         self._grpc_server.add_insecure_port(bind_addr)
         self.logger.debug(f'Start listening on {bind_addr}')
         await self._grpc_server.start()
 
     async def async_run_forever(self):
-        """Block until the GRPC server is terminated """
+        """Block until the GRPC server is terminated"""
         self.connection_pool.start()
         await self._grpc_server.wait_for_termination()
 
@@ -167,10 +193,11 @@ class HeadRuntime(AsyncNewLoopRuntime, ABC):
         :returns: the response request
         """
         try:
-            endpoint = dict(context.invocation_metadata()).get('endpoint')
-            response, metadata = await self._handle_data_request(requests, endpoint)
-            context.set_trailing_metadata(metadata.items())
-            return response
+            with self._summary:
+                endpoint = dict(context.invocation_metadata()).get('endpoint')
+                response, metadata = await self._handle_data_request(requests, endpoint)
+                context.set_trailing_metadata(metadata.items())
+                return response
         except (RuntimeError, Exception) as ex:
             self.logger.error(
                 f'{ex!r}' + f'\n add "--quiet-error" to suppress the exception details'
@@ -178,7 +205,9 @@ class HeadRuntime(AsyncNewLoopRuntime, ABC):
                 else '',
                 exc_info=not self.args.quiet_error,
             )
-            raise
+            requests[0].add_exception(ex, executor=None)
+            context.set_trailing_metadata((('is-error', 'true'),))
+            return requests[0]
 
     async def process_control(self, request: ControlRequest, *args) -> ControlRequest:
         """
@@ -238,8 +267,6 @@ class HeadRuntime(AsyncNewLoopRuntime, ABC):
                 requests, deployment='uses_before'
             )
             requests = [response]
-        elif len(requests) > 1 and self._reduce:
-            requests = [DataRequestHandler.reduce_requests(requests)]
 
         worker_send_tasks = self.connection_pool.send_requests(
             requests=requests,
